'''Package containing simple scripts for plotting objects ``dustpy.Simulation``
or data files.'''

from dustpy.plot.plot import ipanel
from dustpy.plot.plot import panel

__all__ = ["ipanel", "panel"]
__version__ = None
